#############################################################################
##             _____            __ _      _            _                   ##
##            / ____|          / _(_)    (_)          | |                  ##
##           | |     ___   ___| |_ _  ___ _  ___ _ __ | |_ ___             ##
##           | |    / _ \ / _ \  _| |/ __| |/ _ \ '_ \| __/ _ \            ##
##           | |___| (_) |  __/ | | | (__| |  __/ | | | ||  __/            ##
##            \_____\___/ \___|_| |_|\___|_|\___|_| |_|\__\___|            ##
##                                                                         ##
##    Coeficiente Sistemas Ltda. - © 2021. Todos os direitos reservados.   ##
##                                                                         ##
#############################################################################
import logging
from types import SimpleNamespace

from psycopg2 import pool

from asc.core.database.settings import connection_prefs

log = logging.getLogger('psql-connetion')

class PSQLConnectionPooling():
    """
    Implementa um polling de conexões com postgres.
    """
    _registry = {}
    def __init__(self):
        self._connection = pool\
            .ThreadedConnectionPool(**connection_prefs)
        self.active_context = False

    @property
    def connection_used(self):
        """
        método responsável por recuperar todas conexões em uso.
        """
        return self._connection._used

    @property
    def connection_rused(self):
        """
        método responsável por recuperar todas conexões para reutilização.
        """
        return self._connection._rused

    def getpool(self):
        """
        Método responsável por recuperar todas conexões ativas no pool.
        """
        return self._connection._pool

    @classmethod
    def addconnection(cls, row: SimpleNamespace):
        """ método responsável por adicionar a conexão ao registro de classes """
        if not row.name:
            row.name = 'default'

        if row.name not in cls._registry:
            cls._registry[row.name] = row.conn

    @classmethod
    def getconnection(cls, name):
        """ método responsável por obter uma conexão do registro de classes """
        if name in cls._registry:
            return cls._registry[name]
        log.debug(f'recurso {name} não existe no registro')

    @classmethod
    def get_records(cls):
        """ método responsável por obter todos resultados de registro """
        return cls._registry

    @classmethod
    def remove_connection(cls, name):
        """ método responsável por remover uma conexão do registro de classes """
        if name not in cls._registry:
            log.debug(f'Recurso {name} não existe no registro')
            return None
        cls._registry.pop(name)

        log.debug(f'recurso {name} removido com sucesso!')

    @classmethod
    def clear_records(cls):
        """ método responsável por remover todos registros uma
        conexão do registro de classes """
        cls._registry.clear()

    @classmethod
    def len_records(cls):
        """ método responsável por validar o tamanho dos registros de classe """
        return len(cls._registry)

    def getconn(self, name=None):
        """ método responsável por criar uma nova conexão e adicional ao cache.
        :param name: Nome da conexão
        :return: Nova conexão postgres """

        try:
            conn = self._connection.getconn(name)
            self.addconnection(SimpleNamespace(name=name, conn=conn))
            return conn
        except Exception as e:
            print(e)

    def putconn(self, name):
        """ método responsável por adicionar no cache uma conexão para reutilização.
        :param ps_connection: Conexão postgres """
        ps_connection = self.getconnection(name)
        self.remove_connection(name)
        try:
            self._connection.putconn(ps_connection)
        except:
            pass

    def closeall(self):
        """ método responsável por cancelar todas conexão abertas com postgres."""
        self.clear_records()
        self._connection.closeall()

    def connect(self):
        """ métdo responsável por realizar a inicialização do pooling. """
        try:
            self._check_connection()
            self.active_context = True
        except (Exception, psycopg2.DatabaseError) as error:
            self._connection.closeall()

    def _check_connection(self):
        """ método responsável por testar a conexão atual com o postgres. """
        if self._connection:
            return
        raise psycopg2.DatabaseError("Error while connecting to PostgreSQL")

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            if self._connection:
                self.active_context = False
                self.closeall()

    def __repr__(self):
        return (f"{type(self).__name__}("
                f"len_pool='{len(self.getpool())}', "
                f"len_used='{len(self.connection_used)}', "
                f"len_rused={(len(self.connection_rused))}, "
                f"len_records={self.len_records()}, "
                f"context={self.active_context})")

# end-of-file